#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "play.h"

/* Copyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/) *
 * No warranty. License: see file COPYING.                  */

#define USE_FSYNC 0

#if USE_FSYNC
/* Ensure we have no problems with -ansi */
extern int fsync(int);
#endif

#ifdef linux
 #include <sys/soundcard.h>
 #include <sys/ioctl.h>
 #define DEV "/dev/dsp"
#else
 #include <sys/audioio.h>
 #define DEV "/dev/audio"
#endif

#define MIN_TEMPO        1
#define MAX_RATE         64000

/* Maximum length of portamento */
#define PORTA_MAX        32767

#define MAXVOLSET ('z' - 'a' + 1)
#define MAXLOOP 64
#define MAXVOL (255-MAXVOLSET)

/* 

  FIRST NOTES
 
    The language is space- and case insensitive,
      i.e. you may either type 
                  'mtvz fecc#4p8<abpgab>d-epp<a'
               or 'MTVZfeCC#4 p8<ABPGAB>d- ePP<a'.
    However, spaces are not allowed within commands,
      i.e. 'G #6R', 'G# 6R' and 'G#6 R' are not legal.
 
  DEFINITIONS
 
    note   = Case some of: a,b,c,d,e,f,g, possibly followed by +,#,-
             where + and # make sharp, - makes flat.
 
    timing = Effect length = "length" * 2.5 / "tempo" seconds
    tempo  = Default is 125
    length = Default is 6
    volume = Default is 32 (scale: 0=silent, 64=full)
 
  GLOBAL COMMANDS
 
            any local command
            nc...         (declares predefined volume set,
                           where c is index, a-z. Set format:
                           "volume"[:"len"] { , "volume"[:"len"] }
                           If "len" is not specified, 1 is used.
                           If the set ends on :, the last
                           len is assumed to be 255 (eternal).
                           The set decoding ends on the first invalid
                           (non-comma and non-digit) character.)
            t"tempo"      (Changes the tempo.)
            {...}         (Execute simultaneous threads, comma separated)
                          (Each thread consists of local commands, and
                           all the specified threads run simultaneously.)
 
  LOCAL COMMANDS
 
            note[length][r"portamento"]
                          (plays a note of length)
                           "length" may also be period, which
                           specifies current default length*1.5
                           Additionally, "portamento" can be specified.
                           Default "portamento" value is 0.
                           "portamento" specifies the "length"
                           that it takes to slide from a note
                           to another (freq only).
                           In case of only r specified without
                           "portamento" "length", the current
                           effective "length" is used.)
            p[length]     (pauses for "length",
                           period allowed here too.)
            l"length"     (sets default "length")
            >             (increases octave by one)
            <             (decreases octave by one)
            v"volume"     (where "volume" may be either number
                           between 0..64, or predefined volume
                           set index, which is a-z)
            mq            (specifies WAVE_SQ sound envelope)
            mn            (specifies WAVE_NES -"-)
            mt            (specifies WAVE_TRI -"-)
            mr            (specifies WAVE_RAND -"-)
            ms            (specifies WAVE_SINE -"-)
            |             (specify loopstartpoint (vertical wall))
            *"count"      (loop everything between loopstartpoint
                           and this "count" times.)
 
  Notices:
    It is forbidden to change the volume sets while some of them are in use.
    Todo: Check if there is actually need to keep that warning here...
    Loops may be nested until 64th level per thread.
    64 loop nest levels on global level, 64 per single thread.
 
  Thoughts:
    It might be possible to make a speech synthesizer using this program.
    You just have to mix the correct frequencies and sine/rand waves.
    No, I am not going to try :-)
 
  EXAMPLES
  
    l12v16<<<{|cdd+f+f24d+24cdd+<a+>cppp*8,l96>>|p384mt|gp*2*2}
      -- compact emotional song
 
    v16<|<eab>ce<ab>c*2|<eab>cf<ab>c*2|<eab>cf+<ab>c*2|<eab>cf<ab>c*2
      -- something self-unuseful
    
    v64t10l100ms|{<<<<c,>>>>br}*100
      -- test your audio hardware quality.
      -- test your spectrum analyzer software.
 
    mq<na64:1,56:1,48:1,40:1,32:2,24:7,16:2,8:t60l6
    {va|PPedcd12<b>c12dc12<bL12ab24ega36L6ppffga>c18<efga>c12c<b12>c<a90>*2,
     v64<<L24mtaa>ed<gga>e12<a12f>fe<gaaL12>fedc<L48aagaffaL12>fege<}
      -- what is this?
 
 	na22:3,10:nb6:2,2:vbmtva||ec<baeab>c*2|fc<baeab>c*2*22,
 	<mn||bbbbaaaa*2|>cccc<aaaa*2*22,
 	v18msp384|>|<a96f84gr*2>>epl96ge84f12eg84>c12<bp84>c12<bgef84g6p6ec<apppl6*2,
 	mr|p24ep18*88}
	vz fecc#4p8<ab3p.gab>d-epp<ar1
	  -- my first long song with this.
 
*/

static int Rate  = -1;

/* Volume set internal format: 0=length of rest, which is in format: *
 *                             [0]=volume, [1]=len, ...              */
static unsigned char *VolumeSets[MAXVOLSET];
static int Tempo;

static jmp_buf err;

static long ParamRange(long min, long max, const char **s)
{
	long tmp = strtol(*s, (char**)s, 10); /* (*s)--; */
	if(tmp<min || tmp>max)longjmp(err, EINVAL);
	return tmp;
}

#define WAVE_LEN 8192
#define WAVE_SQ    0
#define WAVE_NES   1
#define WAVE_TRI   2
#define WAVE_RAND  3
#define WAVE_SINE  4
#define WAVE_COUNT 5

static int Waves[WAVE_COUNT][WAVE_LEN];
static const float Amplit = 10000;
static const float NumMax = 8;    /* > V64 * 8 makes overflow sound */

/* Len/Rate = seconds - also it's one portamento length unit */
static void Beep(int *Buf,
	int OldNote, int Note, int Portam, int TickPos,
	int Wave, int Len, double *value, int Vol)
{
	static const double powconst = 1.0594630943592952645618252949463417;
	double OldHz = OldNote ? pow(powconst, OldNote+2) : 0.0;
	double Hz    = Note ? pow(powconst,    Note+2) : 0.0;
	int TickPo2  = TickPos+1;
	
	double Hz1 = TickPos>=Portam ? Hz : TickPos * (Hz-OldHz) / Portam+OldHz;
	double Hz2 = TickPo2>=Portam ? Hz : TickPo2 * (Hz-OldHz) / Portam+OldHz;
	
	double delta = Hz1 / Rate;
	double deltadelta = ((Hz2 / Rate) - delta) / Len;
	
	int a;
	
	if(Wave==WAVE_RAND)delta /= 256;
	
	for(a=0; a<Len; a++)
	{
		double dummy;
		*Buf++ += Waves[Wave][(int)(modf(*value, &dummy)*WAVE_LEN)] * Vol / 64;
		*value += delta;
		delta += deltadelta;
	}
}

static void InitWaves(void)
{
	int a;
	float sqr2 = sqrt(2);
	float sqr3 = sqrt(3);
	unsigned long RS=0;
	for(a=0; a<WAVE_LEN; a++)
	{
		float Pos = (float)a / (float)WAVE_LEN;
		
		RS = 1 + RS * 0x8088405UL;  /* Own rand generator */
		
		Waves[WAVE_SQ][a]   = (int)(Amplit * (Pos<=0.5  ? -1 : 1));
		Waves[WAVE_NES][a]  = (int)(Amplit * (Pos<=0.75 ? -1 : 1));
		Waves[WAVE_TRI][a]  = (int)(Amplit * (((Pos<=0.5  ? Pos : (1-Pos)) * 4 - 1) * sqr3));
		Waves[WAVE_RAND][a] = (int)(Amplit * (((RS&65535) / 32768.0 - 1) * sqr3));
		Waves[WAVE_SINE][a] = (int)(Amplit * (sin(Pos*(M_PI+M_PI)) * sqr2));
	}
}

struct Thread
{
	double Value;
	int Length, Volume, Octave, Wave;
	int Note, OldNote, Portam; /* OldNote is for portamento. */
	
	/* While TicksInBuffer decreases, TickPos increases (from 0) */
	int TicksInBuffer, TickPos;
	
	int VolumePos, VolumeLen;
	int LoopPos;
	int LoopCount[MAXLOOP];
	const char *LoopStart[MAXLOOP];
	const char *data;
} MainThread;

static void PushLoop(struct Thread *thr)
{
	if(++thr->LoopPos >= MAXLOOP)longjmp(err, ENOMEM);
}
static void PopLoop(struct Thread *thr)
{
	thr->LoopPos--;
}

static int Unknown = 0; /* Bad way of coding, I know... */
static const char *EvaluateThread(const char *q, int *Buf, int BufSize,
								  struct Thread *thr, int volumedivisor)
{
	while(q && !thr->TicksInBuffer)
	{
		#define Nuotti(_note) { int Len=thr->Length, Note=_note, Portam=0; \
			if(*q=='#' || *q=='+') Note++, q++;	\
			else if(*q=='-') Note--, q++;	\
			if(isdigit((int)*q)) Len=(int)ParamRange(1,INT_MAX, &q); \
			while(*q=='.') q++, Len=Len*3/2; \
			thr->TicksInBuffer=Len;	\
			thr->TickPos   = 0; \
			thr->VolumePos = 1; \
			thr->VolumeLen = 0; \
			if(thr->Note)thr->OldNote = thr->Note; \
			thr->Note = Note; \
			if(*q=='r' && Note!=0) \
			{ \
				q++; \
				Portam=(int)ParamRange(0,PORTA_MAX, &q); \
				if(!Portam)Portam=Len; \
			} \
			thr->Portam = Portam; \
		}
			
		/* Spaces are allowed */
		while(*q==' ' || *q=='\n' || *q=='\r' || *q=='\t')q++;
		
		switch(tolower(*q++)&127)
		{
			case '<': thr->Octave--; break;
			case '>': thr->Octave++; break;
			case 'c': Nuotti(thr->Octave*12);  break;
			case 'd': Nuotti(thr->Octave*12+2); break;
			case 'e': Nuotti(thr->Octave*12+4); break;
			case 'f': Nuotti(thr->Octave*12+5); break;
			case 'g': Nuotti(thr->Octave*12+7); break;
			case 'a': Nuotti(thr->Octave*12+9); break;
			case 'b': Nuotti(thr->Octave*12+11);break;
			case 'p': Nuotti(0); break;
			case '|':
				PushLoop(thr);
				thr->LoopStart[thr->LoopPos] = q;	/* Points to next code */
				thr->LoopCount[thr->LoopPos] = 0;
				break;
			case '*':
			{
				int a;
				if(thr->LoopPos<0)longjmp(err, EINVAL);
				
				a = (int)ParamRange(2,INT_MAX, &q);
				
				switch(thr->LoopCount[thr->LoopPos])
				{
					case 0:	/* New loop */
						thr->LoopCount[thr->LoopPos] = a;
						q = thr->LoopStart[thr->LoopPos];
						break;
					case 2:	/* Finished */
						PopLoop(thr);
						break;
					default:	/* Reloop */
						thr->LoopCount[thr->LoopPos]--;
						q = thr->LoopStart[thr->LoopPos];
				}
				break;
			}
			case 'm':
				switch(tolower(*q++)&127)
				{
					case 'q': thr->Wave = WAVE_SQ; break;
					case 'n': thr->Wave = WAVE_NES; break;
					case 't': thr->Wave = WAVE_TRI; break;
					case 'r': thr->Wave = WAVE_RAND; break;
					case 's': thr->Wave = WAVE_SINE; break;
					default:
						longjmp(err, EINVAL);
				}
				break;
			case 'l': 
				thr->Length = (int)ParamRange(1,INT_MAX, &q);
				break;
			case 'v':
				if(isdigit((int)*q))
					thr->Volume = (int)ParamRange(0,MAXVOL, &q);
				else
				{
					thr->Volume = tolower(*q++);
					if(thr->Volume<'a' || thr->Volume>='z'+MAXVOLSET)
						longjmp(err, EINVAL);
					thr->Volume += MAXVOL-'a'+1;
				}
				break;
			case 0:
			case ',':
			case '}':	/* Closing brace */
				q = NULL;
				break;
			
			case 't':
			case 'n':
			case '{':	/* Opening brace */
				Unknown = 1;
				if(thr==&MainThread)return q-1;
			default:
				longjmp(err, EINVAL);
		}
		
		#undef Nuotti
	}	/* No data available loop */
	
	if(q)	/* Then TicksInBuffer>0 also */
	{
		int Vol = thr->Volume;
		
		thr->TicksInBuffer--;
		
		if(Vol > MAXVOL)
		{
			unsigned char *n = VolumeSets[Vol-MAXVOL-1];
			if(!n)longjmp(err, EINVAL);
			if(!thr->VolumeLen)
			{
				if(thr->VolumePos >= n[0])
					Vol = 0;
				else
				{
					thr->VolumeLen = n[thr->VolumePos+1];
					thr->VolumePos += 2;
				}
			}
			if(Vol)
			{
				/* 255 is considered as eternal */
				if(thr->VolumeLen < 255)thr->VolumeLen--;
				Vol = n[thr->VolumePos-2];
		}	}
		
		Beep(Buf, 
			thr->OldNote,
			thr->Note,
			thr->Portam,
			thr->TickPos,
			thr->Wave,
			BufSize,								
			&thr->Value,
			Vol / volumedivisor);
			
		thr->TickPos++;
		
	}	/* thread still alive */
	return q;
}

#ifdef OPEN_SOUND_SYSTEM
static int Bits;
#endif

static void Writefd(int fd, int *Buf, int BufSize)
{
	static unsigned char Dah[MAX_RATE*5/(MIN_TEMPO+MIN_TEMPO)];
	register int a;
	
	/* Converts signed int (16-bits used) to unsigned char (8-bit) */
	/* In non-OSS case, to signed char (8-bit) */
	
	#ifdef OPEN_SOUND_SYSTEM
	static unsigned short Dah16[2*MAX_RATE*5/(MIN_TEMPO+MIN_TEMPO)];
	if(Bits==16)
	{
		for(a=0; a<BufSize; a++)
		{
			int b = *Buf++ * 32767 / (Amplit*NumMax);
			Dah16[a] = b<-32768 ? -32768 : b>32767 ? 32767 : b;
		}
		write(fd, Dah16, a*2);
	}
	else
	{
		for(a=0; a<BufSize; a++)
		{
			int b = *Buf++ * 127 / (Amplit*NumMax) + 128;
			Dah[a] = b<0 ? 0 : b>255 ? 255 : b;
		}
		write(fd, Dah, a);
	}
	#else
	for(a=0; a<BufSize; a++)
	{
		int b = *Buf++ * 127 / (Amplit*NumMax);
		Dah[a] = b<-128 ? -128 : b>127 ? 127 : b;
	}
	write(fd, Dah, a);	
	#endif
}

static int PlayString(int fd, const char *s)
{
	while(s)
	{
		switch(tolower(*s++)&127)
		{
			case 't':
				Tempo = (int)ParamRange(MIN_TEMPO,INT_MAX, &s);
				break;
			case 'n':
			{
				int p, c = tolower(*s++);
				if(c<'a' || c>='a'+MAXVOLSET)
				{
					errno=EINVAL;
					return -1;
				}
				c -= 'a';
				for(p=1; *s; s++)	/* Post-s++ skips over ',' */
				{
					int vol = (int)ParamRange(0,MAXVOL, &s);
					int len = 1;
					if(*s == ':')
						len = isdigit((int)*++s)
							? (int)ParamRange(1,254, &s)
							: 255;
					VolumeSets[c] = (unsigned char *)realloc(VolumeSets[c], p+2);
					if(!VolumeSets[c])
					{
						errno=ENOMEM;
						return -1;
					}
					VolumeSets[c][0] = p+1;
					VolumeSets[c][p+1] = len;
					VolumeSets[c][p] = vol;
					p += 2;
					if(*s != ',')break;
				}
				break;
			}
			case '{':	/* Opening brace */
			{
				/* The thread data and -type */
				struct Thread *threads;
				
				int threadcount, new, threadsleft;
				const char *q;
				for(q=--s, threadcount=1; *++q != '}'; ) /* Closing brace */
					if(*q == ',')threadcount++;
					
				threads = (struct Thread *)calloc(threadcount, sizeof(struct Thread));
				if(!threads)
				{
					errno=ENOMEM;
					return -1;
				}
				
				for(new=1, threadcount=threadsleft=0; ;)
				{
					s++;
					if(new)
					{
						threads[threadcount] = MainThread;
						threads[threadcount].data = s;
						threads[threadcount].LoopPos = -1;
						threadcount++;
						threadsleft++;
						new = 0;
					}
					if(*s == 125)break;	/* closing brace */
					if(*s == ',')new = 1;
				}
				s++;	/* Skip brace */
				
				while(threadsleft)
				{
					int a;
					int BufSize = Rate * 5 / (Tempo+Tempo);
					int *Buf;
					
					Buf = (int *)calloc((sizeof(int)), BufSize);
					if(!Buf)
					{
						free(threads);
						errno = ENOMEM;
						return -1;
					}
					
					for(a=0; a<threadcount; a++)
						if(threads[a].data)
						{
							if(!(threads[a].data = 
								EvaluateThread(
									threads[a].data, Buf, BufSize,
									&threads[a], 
									#if 0
									threadcount
									#else
									1	/* This causes too easily *
										 * volume overflows.      */
									#endif
									)))
								threadsleft--;
						}
					
					Writefd(fd, Buf, BufSize);
					
					free(Buf);
				}
				free(threads);
				break;
			}	/* Opening brace case */
			
			default:
			{
				int BufSize = Rate * 5 / (Tempo+Tempo);
				int *Buf;
				
				Buf = (int *)malloc(BufSize * (sizeof(int)));
				if(!Buf)
				{
					errno = ENOMEM;
					return -1;
				}
				
				for(s--;;)
				{
					memset(Buf, 0, BufSize*(sizeof(int)));
					s = EvaluateThread(s, Buf, BufSize, &MainThread, 1);
					if(Unknown)
					{	
						Unknown = 0;
						break;
					}
					Writefd(fd, Buf, BufSize);
					
					if(!MainThread.TicksInBuffer)break;
				}
				
				free(Buf);
			}
		}
	}
	return 0;
}

static int fd;

int Play(const char *s)
{
	int a;
	Tempo = 125;
	
	MainThread.TicksInBuffer = 0;
	MainThread.Length = 6;
	MainThread.Volume = 32;
	MainThread.Octave = 9;
	MainThread.Wave   = WAVE_NES;
	MainThread.Value  = 0.0;
	MainThread.LoopPos= -1;
	
	a=setjmp(err);
	if(a)
	{
		errno=a;
		return -1;
	}	
	a = PlayString(fd, s);
	
#if USE_FSYNC
	fsync(fd);
#endif
	
	return a;
}

int InitPlay(int bits, int rate)
{
	Rate = rate;
	
	#if 0
	if(Stdout)
		fd = fileno(stdout);
	else
	#endif
	{
	#if 1
		int pip[2], pidi;
		if(pipe(pip) < 0)return -1;
		Bits = bits==8 ? 8 : 16;
		pidi = fork();
		if(pidi < 0)return -1;
		if(!pidi)
		{
			char rateparm[10];
			
			close(pip[1]);
			dup2(pip[0], 0);
			
			sprintf(rateparm, "%d", Rate);
			
			if(Bits == 8)
				execl("/usr/local/bin/esdcat", "esdcat", "-m", "-r", rateparm, "-s", "127.0.0.1", "-b", NULL);
			else
				execl("/usr/local/bin/esdcat", "esdcat", "-m", "-r", rateparm, "-s", "127.0.0.1", NULL);
			
			fprintf(stderr, "exec failed??\n");
		}
		close(pip[0]);
		fd = pip[1];
	#else
		fd = open(DEV, O_WRONLY);
		if(fd < 0)return -1;
			
		#ifdef OPEN_SOUND_SYSTEM
		Bits=bits;
		if(ioctl(fd, SNDCTL_DSP_SPEED, &Rate) < 0
		|| ioctl(fd, SNDCTL_DSP_SAMPLESIZE, &Bits) < 0)
		{
			close(fd);
			return(fd=-1);
		}
		#else
		{
			audio_info_t ainfo;
			AUDIO_INITINFO(&ainfo);
			ainfo.play.sample_rate = Rate;
			ainfo.play.channels = 1;
			ainfo.play.precision= 8;
			ainfo.play.encoding = AUDIO_ENCODING_LINEAR;
			ainfo.play.gain = -1; /* 64; */
			ainfo.play.port = AUDIO_SPEAKER;
			ainfo.play.balance = AUDIO_MID_BALANCE;
			ainfo.play.buffer_size = 8180;
			if(ioctl(fd, AUDIO_SETINFO, &ainfo) == -1)
			{
				close(fd);
				return(fd=-1);
			}
		}
		#endif
	#endif
	}
	
	InitWaves();
	
	return 0;
}

void DonePlay(void)
{
	if(fd>=0)
	{
		close(fd);
		fd=-1;
	}
}
